// VGM player for the NeoGeo Pocket
// Simple playback example / C version
// TLCS900H part
// /Mic, 2012

#include "ngpc.h"
#include "carthdr.h"
#include "library.h"
#include <stdlib.h>
#include <stdio.h>
#include "vgmdriv.h"	// VGM playback driver for the Z80
#include "song.h"		// VGM data

typedef volatile u8 vu8;
typedef volatile u16 vu16;

#define memcpy(dst, src, len) BlockCopy(dst, src, len)

#define Z80_OFF_SOUND_OFF() SOUNDCPU_CTRL = 0xAAAA
#define Z80_ON_SOUND_ON()	SOUNDCPU_CTRL = 0x5555

#define VGM_BUFFER_READ_POS	Z80_COMM

// Variables in Z80 RAM
#define VGM_DATA_BUFFER 	(*(vu8*)0x7300)
#define VGM_RESTART			(*(vu8*)0x7080)


typedef struct {
	u32 signature;
	u32 eofOffset;
	u32 version;
	u32 sn76489Clock;
	u32 ym2413Clock;
	u32 gd3Offset;
	u32 totNumSamples;
	u32 loopOffset;
	u32 loopNumSamples;
	u32 rate;
	u32 sn76489Feedback;
	u32 y2612Clock;
	u32 ym2151Clock;
	u32 dataOffset;
} VgmHeader_t;


VgmHeader_t *vgmHeader;
vu8 *vgmPtr, *vgmSrc;
u8 vgmDstOffs;
u8 rstat;
u16 vgmHeaderSize;

void __interrupt update_vgm(void);


void main(void) {
    rstat = USR_BOOT;
    InitNGPC();
    SysSetSystemFont();

	vgmSrc			= 0;
	Z80_OFF_SOUND_OFF();

	// Hook our VGM update function to the vblank
	// interrupt
	VBL_INT 		= update_vgm;

	vgmPtr			= (vu8*)SONG;
	vgmHeader 		= (VgmHeader_t*)vgmPtr;
	vgmHeaderSize 	= 0;

	if ((u16)vgmHeader->version > 0x150) {
		vgmHeaderSize = vgmHeader->dataOffset + 0x34;
	} else {
		vgmHeaderSize = 0x40;
	}

	// Copy the playback driver to Z80 RAM
	memcpy(Z80_RAM, VGM_SMS_Z80_DRIVER, 0x1000);

	// Fill the VGM data buffer in Z80 RAM with the first
	// 256 bytes of VGM data
	memcpy(&VGM_DATA_BUFFER, &vgmPtr[vgmHeaderSize], 256);

	//VGM_BUFFER_READ_POS = 0;
	VGM_RESTART 		= 0;

	vgmSrc = vgmPtr + vgmHeaderSize + 256;
	vgmDstOffs = 0;

	Z80_ON_SOUND_ON();

	while (1) {}
}



// Checks if more VGM data needs to be copied to the
// shared RAM area.
//
void __interrupt update_vgm(void) {
	u16 i, loopOffs;
	u8 bytesToCopy;
	vu8 *dst = &VGM_DATA_BUFFER;

	// Clear the watchdog
	WATCHDOG = WATCHDOG_CLEAR;

	// Check for power down
	if (USR_SHUTDOWN)
	{
		SysShutdown();
		while (1);
	}

	VBCounter++;

	if (vgmSrc) {
		if (VGM_RESTART) {
			// The end of the song has been reached. Calculate the
			// loop point and start reading data from there
			loopOffs = vgmHeader->loopOffset;
			if (loopOffs == 0) {
				loopOffs = vgmHeaderSize - 0x1C;
			}
			vgmSrc = vgmPtr + 0x1C + loopOffs;

			// Copy data to where the Z80 will be reading from next
			vgmDstOffs = VGM_BUFFER_READ_POS + 1;
			for (i = 0; i < 256; i++) {
				dst[vgmDstOffs++] = *vgmSrc++;
			}
			// Tell the Z80 to resume playback
			VGM_RESTART = 0;

		} else {
			// Send some more data if the Z80 is getting close to having
			// consumed all VGM data written to the buffer
			bytesToCopy = VGM_BUFFER_READ_POS - vgmDstOffs;
			bytesToCopy >>= 1;
			while (bytesToCopy--) {
				dst[vgmDstOffs++] = *vgmSrc++;
			}
		}
	}
}


